import time
from datetime import date, datetime, timedelta
from typing import Optional

from abc import ABC, abstractmethod
from django.utils import timezone

from greg.models import PersonRole, Notification, ScheduleTask


class BaseSchedule(ABC):
    """
    Provides common methods for tasks
    """

    task_object: ScheduleTask

    def __init__(self, name: str):
        self.task_object = ScheduleTask.objects.get(name=name)

    def run(self):
        self._run_internal()
        self.task_object.last_completed = timezone.now()
        self.task_object.save()

    @abstractmethod
    def _run_internal(self):
        pass

    def get_last_run(self) -> Optional[datetime]:
        self.task_object.refresh_from_db()
        return self.task_object.last_completed

    def _store_notification(
        self, identifier, object_type, operation, **kwargs
    ) -> Notification:
        return Notification.objects.create(
            identifier=identifier,
            object_type=object_type,
            operation=operation,
            issued_at=int(time.time()),
            meta=kwargs,
        )


class ExpiringRolesNotification(BaseSchedule):
    """
    This task does a simple check for roles that will expire in 30 days
    and creates entries in the notification table for them.

    There should be an entry in the ScheduleTask-table with name role_expiration
    before this task is run.

    Some assumptions that are made:

    - The task will be run every day, it does
    not keep track of which roles it has or has not notified, it only
    looks at which roles will expire in exactly 30 days

    - If there are roles created that expire in less than 30 days, no
    notification is necessary
    """

    EXPIRATION_THRESHOLD_DAYS = 30

    def __init__(self):
        # Will raise exception if there is not exactly one result
        super().__init__("role_expiration")

    def _run_internal(self):
        last_run = self.get_last_run()
        # Only run once per day
        if last_run is None or last_run.date() != date.today():
            check_date = datetime.today() + timedelta(
                days=self.EXPIRATION_THRESHOLD_DAYS
            )
            self.__get_roles_about_to_expire(check_date)

    def __get_roles_about_to_expire(self, end_date: date):
        roles_about_to_expire = PersonRole.objects.filter(end_date=end_date)

        for person_role in roles_about_to_expire:
            meta = {"person_id": person_role.person.id, "role_id": person_role.role.id}
            self._store_notification(
                identifier=person_role.id,
                object_type="PersonRole",
                operation="expire_reminder",
                **meta
            )
