from datetime import (
    datetime,
    timedelta,
)
from functools import partial

import pytest

from greg.models import (
    OrganizationalUnit,
    Person,
    PersonIdentity,
    PersonRole,
    Role,
    Sponsor,
)

person_role_with = partial(
    PersonRole.objects.create,
    start_date="2020-03-05",
    end_date="2020-06-10",
    contact_person_unit="Contact Person",
    available_in_search=True,
)


@pytest.fixture
def role_foo() -> Role:
    return Role.objects.create(type="role_foo", name_en="Role Foo")


@pytest.fixture
def role_bar() -> Role:
    return Role.objects.create(type="role_bar", name_en="Role Bar")


@pytest.fixture
def person(role_foo: Role, role_bar: Role) -> Person:
    person = Person.objects.create(
        first_name="Test",
        last_name="Tester",
        date_of_birth="2000-01-27",
        email="test@example.org",
        mobile_phone="123456789",
    )

    ou = OrganizationalUnit.objects.create(orgreg_id="12345", name_en="Test unit")
    person_role_with(
        person=person,
        role=role_foo,
        unit=ou,
        registered_by=Sponsor.objects.create(feide_id="foosponsor@uio.no"),
    )
    person_role_with(
        person=person,
        role=role_bar,
        unit=ou,
        registered_by=Sponsor.objects.create(feide_id="barsponsor@uio.no"),
    )

    return person


@pytest.fixture
def a_year_ago() -> datetime:
    return datetime.now() - timedelta(days=365)


@pytest.fixture
def a_year_into_future() -> datetime:
    return datetime.now() + timedelta(days=365)


@pytest.fixture
def feide_id(a_year_ago: datetime) -> PersonIdentity:
    return PersonIdentity(
        type=PersonIdentity.IdentityType.FEIDE_ID,
        verified_when=a_year_ago,
    )


@pytest.fixture
def passport(a_year_ago: datetime) -> PersonIdentity:
    return PersonIdentity(
        type=PersonIdentity.IdentityType.PASSPORT,
        verified_when=a_year_ago,
    )


@pytest.fixture
def unverified_passport() -> PersonIdentity:
    return PersonIdentity(type=PersonIdentity.IdentityType.PASSPORT)


@pytest.fixture
def future_identity(a_year_into_future: datetime) -> PersonIdentity:
    return PersonIdentity(
        type=PersonIdentity.IdentityType.NATIONAL_ID_CARD,
        verified_when=a_year_into_future,
    )


@pytest.fixture
def feide_verified(person: Person, feide_id: PersonIdentity) -> Person:
    person.identities.add(feide_id, bulk=False)
    return person


@pytest.mark.django_db
def test_add_multiple_roles_to_person(person, role_foo, role_bar):
    person_roles = person.roles.all()
    assert len(person_roles) == 2
    assert role_foo in person_roles
    assert role_bar in person_roles


@pytest.mark.django_db
def test_identities(person: Person, feide_id: PersonIdentity, passport: PersonIdentity):
    person.identities.add(feide_id, bulk=False)
    assert list(person.identities.all()) == [feide_id]
    person.identities.add(passport, bulk=False)
    assert list(person.identities.all()) == [feide_id, passport]


@pytest.mark.django_db
def test_is_registered_incomplete(person):
    assert person.registration_completed_date is None
    assert not person.is_registered


@pytest.mark.django_db
def test_is_registered_completed_in_past(person, a_year_ago):
    person.registration_completed_date = a_year_ago
    assert person.is_registered


@pytest.mark.django_db
def test_is_registered_completed_in_future(person, a_year_into_future):
    person.registration_completed_date = a_year_into_future
    assert not person.is_registered


@pytest.mark.django_db
@pytest.mark.parametrize("identity_type", PersonIdentity.IdentityType)
def test_is_verified(identity_type, a_year_ago, person):
    identity = PersonIdentity(type=identity_type, verified_when=a_year_ago)
    person.identities.add(identity, bulk=False)
    assert person.is_verified


@pytest.mark.django_db
def test_is_verified_multiple_identities(person, feide_id, passport):
    person.identities.add(feide_id, passport, bulk=False)
    assert person.is_verified


@pytest.mark.django_db
def is_verified_when_identity_is_unverified(person, unverified_passport):
    person.identities.add(unverified_passport, bulk=False)
    assert not person.is_verified


@pytest.mark.django_db
def is_verified_mixed_verified_and_unverified_identities(
    person,
    feide_id,
    unverified_passport,
    future_identity,
):
    person.identities.add(feide_id, unverified_passport, future_identity, bulk=False)
    assert person.is_verified


@pytest.mark.django_db
def is_verified_in_future(person, future_identity):
    person.identities.add(future_identity, bulk=False)
    assert not person.is_verified
