from datetime import datetime, timedelta

import pytest
from django.db.models import Q
from django.utils import timezone

from greg.models import (
    ScheduleTask,
    Role,
    Person,
    OrganizationalUnit,
    PersonRole,
    Notification,
    Sponsor,
)
from greg.schedule import ExpiringRolesNotification


@pytest.fixture
def role_task():
    ScheduleTask.objects.create(
        name="role_expiration", last_completed=timezone.now() - timedelta(days=1)
    )


@pytest.fixture
def role_bar() -> Role:
    return Role.objects.create(type="role_bar", name_en="Role Bar")


@pytest.fixture
def person(role_bar: Role) -> Person:
    return Person.objects.create(
        first_name="Test",
        last_name="Tester",
        date_of_birth="2000-01-27",
        email="test@example.org",
        mobile_phone="123456789",
    )


@pytest.fixture
def organizational_unit() -> OrganizationalUnit:
    return OrganizationalUnit.objects.create(orgreg_id="12345", name_en="Test unit")


@pytest.fixture
def sponsor() -> Sponsor:
    return Sponsor.objects.create(feide_id="sponsor@example.org")


@pytest.fixture
def person_role(
    person: Person,
    role_bar: Role,
    organizational_unit: OrganizationalUnit,
    sponsor: Sponsor,
) -> PersonRole:
    return PersonRole.objects.create(
        person=person,
        role=role_bar,
        unit=organizational_unit,
        start_date="2020-03-05",
        end_date=datetime.today() + timedelta(days=30),
        contact_person_unit="Contact Person",
        available_in_search=True,
        registered_by=sponsor,
    )


@pytest.mark.django_db
def test_role_picked_up(role_task: ScheduleTask, person_role: PersonRole):
    role_notification = ExpiringRolesNotification()
    assert len(Notification.objects.filter(~Q(operation="add"))) == 0
    role_notification.run()
    notification = Notification.objects.get(operation="expire_reminder")
    assert notification.identifier == person_role.id
    role_notification.run()
    notifications = Notification.objects.filter(operation="expire_reminder")
    assert len(notifications) == 1


@pytest.mark.django_db
def test_no_notification_for_role_not_about_to_expire(
    role_task: ScheduleTask, person_role: PersonRole
):
    person_role.end_date = datetime.today() + timedelta(days=31)
    person_role.save()
    role_notification = ExpiringRolesNotification()
    assert len(Notification.objects.filter(operation="expire_reminder")) == 0
    # Role should not be picked up since it expires in 31 days
    role_notification.run()
    notifications = Notification.objects.filter(operation="expire_reminder")
    assert len(notifications) == 0
